#!/bin/bash
########################################################################
# build_tbb.sh
#
# Build tbb and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_03_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=tbb
origpkgver=v2020_2
pkgver=${origpkgver}a # Append letters for local patches.
ssibuildshims_version=v2_03_02
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=oneTBB-${pkgdotver}
pkgtarfile=${pkgtarname}.tar.gz
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,e21,c7} \
  --btype={debug,prof} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics TRACK_PROGRESS IN_SOURCE || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
#
# e.g.
#   setup_cmake --latest || exit
# OR
#   setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" \
  tbb.patch \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

case ${cqual} in
  e19|e20) compiler=gcc; cxxstd=17;;
  e21) compiler=gcc; cxxstd=2a;;
  c7) compiler=clang; cxxstd=17;;
  *) ssi_die -I "accepted unknown cqual=${cqual}"
esac
case ${btype} in
  debug) reltype=debug; suffix=_debug;;
  prof) reltype=release;;
  *) ssi_die -I "accepted unknown btype=${btype}"
esac

##################
# Configure parallelism. *C*
ncores=$(ncores)
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}/${pkgtarname}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*
# Build. *C*

${trace} make -C src -j ${ncores} \
  compiler=${compiler} stdver=c++${cxxstd} \
  tbb{,malloc}_${reltype} || \
  ssi_die "build failed"
##################

##################
# Test. *C*
##################

##################
# Install. *C*

ssi_info "A la carte install"
${trace} mv -v include "${pkgdir}/" || \
  ssi_die "failed to install include"
${trace} mkdir -p "${pkgdir}/lib" || \
  ssi_die "failed to create ${pkgdir}/lib"
if [ "${OS1}" = Darwin ]; then
  ext=.dylib
  ${trace} mv -v build/macos/lib*${ext} "${pkgdir}/lib/" || \
    ssi_die "failed to install MacOS shared libraries"
else
  ext=.so
  ${trace} mv -v build/lin*/lib*${ext}* "${pkgdir}/lib/" || \
    ssi_die "failed to install Linux shared libraries"
  if [ -d build/mic_icc_${reltype} ]; then
    ${trace} mkdir "${pkgdir}/lib-mic"
    ${trace} mv -v build/mic_icc_${reltype}/libtbb*.so* "${pkgdir}/lib-mic" || \
      ssi_die "failed to install Intel Xeon Phi offload libraries"
  fi
fi
${trace} rm -rf "${pkgdir}/doc" && ${trace} mv -v doc "${pkgdir}/" || \
  ssi_die "failed to install docs"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
ssi_info "clean source"
make compiler=${compiler} -j ${ncores} clean

ssi_info "create soft links for libraries"
# 1. Replace linker stubs with soft links to avoid confusing CMake when
# trying to link to these libraries (not for OS X).
#
# 2. For debug, make soft links without _debug in the name for convenience.
for dir in "${pkgdir}/lib" "${pkgdir}/lib-mic"; do
  [[ -d "${dir}" ]] || continue
  pushd "${dir}" >/dev/null
  for lib in libtbb{,malloc{,_proxy}}"${suffix}${ext}"; do
    soname=$(grep -qI INPUT "${lib}" && cat "${lib}" | sed -e 's&^INPUT (\(.*\))$&\1&')
    [[ -n "${soname}" ]] && ln -svf "${soname}" "${lib}" # 1.
    [[ -n "${suffix}" ]] && ln -svf "${soname:-${lib}}" "${lib/${suffix}/}" # 2.
  done
  popd >/dev/null
done

ssi_info "install source"
${trace} rm -rf "${pkgtop}/source" && \
  ${trace} mv -v src "${pkgdir}/source" || \
  ssi_die "unable to install source"

ssi_info "install and configure CMake Config files"
# cmake configure files
cp -pr "${patchdir}/${package}" "${pkgdir}/lib/"
# Ensure version is set correctly.
config_files=("${pkgdir}/lib/tbb/cmake/tbb-config"{,-version}".cmake")
for config_file in "${config_files[@]}"; do
  perlfilt -wpi'' \
     -e 's&\%\%\%PKGVER\%\%\%&'"${pkgver}"'&;
         s&\%\%\%PKGDOTVER\%\%\%&'"${pkgdotver}"'&;' \
     "${config_file}" || ssi_die "unable to update version in ${config_file}"
done
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
